from __future__ import annotations

import re
from dataclasses import dataclass, asdict
from http.client import responses
from json import JSONDecodeError
from typing import Optional, TYPE_CHECKING

import httpx

from .enums import Message as ResponseMessage
from .enums import Status as ResponseStatus
from .exceptions import (
    HTTPError,
    BadRequest,
    TokenInvalid,
    InvalidCredentials,
)
from .logging import logger

if TYPE_CHECKING:
    from .base import EskizSMSBase

BASE_URL = "https://notify.eskiz.uz/api"
API_VERSION_RE = re.compile("API version: ([0-9.]+)")


def _url(path: str):
    return BASE_URL + path


@dataclass
class _Response:
    status_code: int
    data: dict
    token_invalid: bool = False


@dataclass
class _Request:
    method: str
    url: str
    data: dict = None
    headers: dict = None


class BaseRequest:

    @staticmethod
    def _prepare_request(method: str, path: str, data: dict = None, headers: dict = None):
        return _Request(method, _url(path), data, headers)

    @staticmethod
    def _exception(_response: _Response):
        status = _response.data.get('status')
        message = _response.data.get('message') or responses[_response.status_code]
        if status == ResponseStatus.TOKEN_INVALID:
            return TokenInvalid(
                message=message,
                status=status,
                status_code=_response.status_code
            )
        if message == ResponseMessage.INVALID_CREDENTIALS:
            return InvalidCredentials(message="Invalid credentials", status_code=_response.status_code)
        return BadRequest(
            message=message,
            status=status,
            status_code=_response.status_code
        )

    @staticmethod
    def _get_authorization_header(token):
        return {
            "Authorization": f"Bearer {token}"
        }

    def _request(self, _request: _Request):
        try:
            with httpx.Client() as client:
                return self._check_response(client.request(**asdict(_request)))
        except httpx.HTTPError as e:
            raise HTTPError(message=str(e))

    async def _a_request(self, _request: _Request):
        try:
            async with httpx.AsyncClient() as client:
                return self._check_response(await client.request(**asdict(_request)))
        except httpx.HTTPError as e:
            raise HTTPError(message=str(e))

    def _check_response(self, r: httpx.Response) -> _Response:
        response: Optional[_Response] = None
        try:
            response = _Response(status_code=r.status_code, data=r.json())
        except JSONDecodeError:
            if r.status_code == 200:
                api_version = API_VERSION_RE.search(r.text)
                if api_version:
                    response = _Response(status_code=r.status_code, data={'api_version': api_version.groups()[0]})

        if response is None:
            response = _Response(status_code=r.status_code, data={'message': responses[r.status_code]})

        logger.debug(f"Eskiz request status_code={response.status_code} body={response.data}")

        if response.status_code == 401:
            if response.data.get('status') == ResponseStatus.TOKEN_INVALID:
                response.token_invalid = True
                return response

        if response.status_code not in [200, 201]:
            raise self._exception(response)

        return response


class Request(BaseRequest):
    def __init__(self, eskiz: EskizSMSBase):
        self._eskiz = eskiz

    def __call__(self, method: str, path: str, payload: dict = None):
        _request = self._prepare_request(
            method,
            path,
            data=self._prepare_payload(payload)
        )
        if getattr(self._eskiz, 'is_async', False):  # noqa
            return self.async_request(_request)
        return self.request(_request)

    async def async_request(self, _request: _Request) -> dict:
        _request.headers = self._get_authorization_header(await self._eskiz.token.get())
        response = await self._a_request(_request)
        if response.token_invalid and self._eskiz.token.auto_update:
            logger.debug("Refreshing the token")
            _request.headers = self._get_authorization_header(await self._eskiz.token.get(get_new=True))
            response = await self._a_request(_request)
        if response.status_code not in [200, 201]:
            raise self._exception(response)
        return response.data

    def request(self, _request: _Request) -> dict:
        _request.headers = self._get_authorization_header(self._eskiz.token.get())
        response = self._request(_request)
        if response.token_invalid and self._eskiz.token.auto_update:
            logger.debug("Refreshing the token")
            _request.headers = self._get_authorization_header(self._eskiz.token.get(get_new=True))
            response = self._request(_request)
        if response.status_code not in [200, 201]:
            raise self._exception(response)
        return response.data

    @staticmethod
    def _prepare_payload(payload: dict):
        payload = payload or {}
        if 'from_whom' in payload:
            payload['from'] = payload.pop('from_whom')
        if 'mobile_phone' in payload:
            payload['mobile_phone'] = payload['mobile_phone'].replace("+", "").replace(" ", "")
        return payload

    def post(self, path: str, payload: dict = None):
        return self("POST", path, payload)

    def put(self, path: str, payload: dict = None):
        return self("PUT", path, payload)

    def get(self, path: str, payload: Optional[dict] = None):
        return self("GET", path, payload)

    def delete(self, path: str, payload: dict = None):
        return self("DELETE", path, payload)

    def patch(self, path: str, payload: dict = None):
        return self("PATCH", path, payload)



import re
from typing import Optional, List

from eskiz_sms.request import Request
from .exceptions import InvalidCallbackUrl
from .token import Token
from .types import User, Contact, Response


class Meta(type):
    def __new__(mcs, name, bases, dct, async_=False):
        x = super().__new__(mcs, name, bases, dct)
        setattr(x, 'is_async', async_)
        return x


class EskizSMSBase(metaclass=Meta):
    __slots__ = (
        "token",
        "_user",
        "callback_url",
        "is_async",
        "_request",
    )

    def __init__(
            self,
            email: str,
            password: str,
            callback_url: str = None,
            save_token: bool = False,
            env_file_path: str = None,
            auto_update_token=True,
    ):

        if callback_url is not None:
            self._validate_callback_url(callback_url)
        self.callback_url = callback_url

        self.token = Token(
            email,
            password,
            save_token=save_token,
            env_file_path=env_file_path,
            auto_update=auto_update_token,
            is_async=getattr(self, 'is_async')
        )
        self._request = Request(self)
        self._user: Optional[User] = None

    @staticmethod
    def _validate_callback_url(url):
        if url_validator(url) is False:
            raise InvalidCallbackUrl(message="Invalid callback url")

    def _get_callback_url(self, callback_url: str = None):
        if callback_url is not None:
            self._validate_callback_url(callback_url)
            return callback_url
        return self.callback_url

    @property
    def user(self) -> Optional[User]:
        raise NotImplementedError

    def _user_data(self) -> Optional[User]:
        raise NotImplementedError

    def add_contact(self, name: str, email: str, group: str, mobile_phone: str) -> Contact:
        raise NotImplementedError

    def update_contact(self, contact_id: int, name: str, group: str, mobile_phone: str) -> Optional[Contact]:
        raise NotImplementedError

    def get_contact(self, contact_id: int, raise_exception=False) -> Optional[Contact]:
        raise NotImplementedError

    def delete_contact(self, contact_id: int) -> Response:
        raise NotImplementedError

    def send_sms(self, mobile_phone: str, message: str, from_whom: str = '4546',
                 callback_url: str = None) -> Response:
        """
        :param mobile_phone: Phone number without plus sign
        :param message: Message to send
        :param from_whom: To use a nickname, you need to change the field to your own
        :param callback_url: This is an optional field that is used to automatically receive SMS status from the server.
            Specify the callback URL where you will receive POST data in the following format:
            {"message_id": "4385062", "user_sms_id": "your_id_here", "country": "UZ",
            "phone_number": "998991234567", "sms_count": "1",
            "status" : "DELIVER", "status_date": "2021-04-02 00:39:36"}
        :return: Response
        """
        raise NotImplementedError

    def send_global_sms(self, mobile_phone: str, message: str, country_code: str,
                        callback_url: str = None, unicode: str = "0") -> Response:
        """
        :param mobile_phone: Phone number without plus sign
        :param message: Message to send
        :param country_code: e.g. 'US'
        :param callback_url: Pass a callback url to get notified when the message is sent
        :param unicode: Default is 0, pass 1 if you want to send cyrillic message
        :return:
        """

        raise NotImplementedError

    def send_batch(self, *, messages: List[dict], from_whom: str = "4546", dispatch_id: int) -> Response:
        """
        :param messages: List of messages to send.
            [{"user_sms_id":"sms1","to": 998998046210, "text": "eto test"}]
        :param from_whom: 4546
        :param dispatch_id:
        :returns: Response
        :rtype: eskiz_sms.types.Response
        """
        raise NotImplementedError

    def get_user_messages(self, from_date: str, to_date: str) -> Response:
        raise NotImplementedError

    def get_user_messages_by_dispatch(self, dispatch_id: int) -> Response:
        raise NotImplementedError

    def get_dispatch_status(self, dispatch_id: int) -> Response:
        raise NotImplementedError

    def create_template(self, name: str, text: str) -> Response:
        raise NotImplementedError

    def update_template(self, template_id: int, name: str, text: str) -> Response:
        raise NotImplementedError

    def get_template(self, template_id: int) -> Response:
        raise NotImplementedError

    def get_templates(self) -> Response:
        raise NotImplementedError

    def totals(self, year: int) -> Response:
        raise NotImplementedError

    def get_limit(self) -> Response:
        raise NotImplementedError


URL_RE = re.compile(
    r"[(http(s)?):\/\/(www\.)?a-zA-Z0-9@:%._\+~#=]"  # noqa
    r"{2,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)"  # noqa
)


def url_validator(url: str):
    return bool(URL_RE.search(url))



class EskizException(Exception):
    def __init__(self, message=None, status=None, status_code: int = None):
        self.status = status
        self.message = message
        message = str(message) if message else ''
        if status:
            message += f"; status={status}"
        if status_code:
            message += f"; status_code={status_code}"
        super(EskizException, self).__init__(message.strip())


class BadRequest(EskizException):
    pass


class InvalidCredentials(EskizException):
    pass


class InvalidCallbackUrl(EskizException):
    pass


class TokenBlackListed(EskizException):
    pass


class TokenInvalid(EskizException):
    pass


class ContactNotFound(EskizException):
    pass


class HTTPError(EskizException):
    pass



from dataclasses import dataclass
from datetime import datetime
from typing import Union, Optional


@dataclass
class Response:
    id: Optional[str] = None
    status: Optional[str] = None
    data: Optional[Union[dict, list]] = None
    message: Optional[Union[str, dict]] = None


@dataclass
class User:
    id: Optional[int] = None
    name: Optional[str] = None
    email: Optional[str] = None
    role: Optional[str] = None
    api_token: Optional[str] = None
    status: Optional[str] = None
    sms_api_login: Optional[str] = None
    sms_api_password: Optional[str] = None
    uz_price: Optional[int] = None
    ucell_price: Optional[int] = None
    test_ucell_price: Optional[int] = None
    balance: Optional[int] = None
    is_vip: Optional[bool] = None
    host: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class Contact:
    id: Optional[int] = None
    user_id: Optional[int] = None
    group: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    mobile_phone: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class ContactCreated:
    contact_id: int



from typing import Optional, List

from .base import EskizSMSBase
from .exceptions import ContactNotFound
from .types import User, Contact, Response, ContactCreated


class EskizSMS(EskizSMSBase):
    @property
    def user(self) -> Optional[User]:
        self._user = self._user_data()
        return self._user

    def _user_data(self) -> Optional[User]:
        response = self._request.get("/auth/user")
        return User(**response)

    def add_contact(self, name: str, email: str, group: str, mobile_phone: str) -> ContactCreated:
        response = self._request.post(
            "/contact",
            payload={
                "name": name,
                "email": email,
                "group": group,
                "mobile_phone": str(mobile_phone),
            })
        return ContactCreated(response['data'])

    def update_contact(self, contact_id: int, name: str, group: str, mobile_phone: str) -> Optional[Contact]:
        response = self._request.put(
            f"/contact/{contact_id}",
            payload={
                "name": name,
                "group": group,
                "mobile_phone": str(mobile_phone),
            })
        if response and isinstance(response, list):
            return Contact(**response[0])

    def get_contact(self, contact_id: int, raise_exception=False) -> Optional[Contact]:
        response = self._request.get(f"/contact/{contact_id}")
        if not response:
            if raise_exception:
                raise ContactNotFound
            return None
        return Contact(**response[0])

    def delete_contact(self, contact_id: int) -> Response:
        response = self._request.delete(f"/contact/{contact_id}")
        return Response(**response)

    def send_sms(self, mobile_phone: str, message: str, from_whom: str = '4546',
                 callback_url: str = None) -> Response:

        payload = {
            "mobile_phone": str(mobile_phone),
            "message": message,
            "from_whom": from_whom,
        }
        callback_url = self._get_callback_url(callback_url)
        if callback_url:
            payload['callback_url'] = callback_url
        return Response(**self._request.post("/message/sms/send", payload=payload))

    def send_global_sms(self, mobile_phone: str, message: str, country_code: str,
                        callback_url: str = None, unicode: str = "0") -> Response:
        payload = {
            "mobile_phone": str(mobile_phone),
            "message": message,
            "country_code": country_code,
            "unicode": unicode
        }
        callback_url = self._get_callback_url(callback_url)
        if callback_url:
            payload['callback_url'] = callback_url
        return Response(**self._request.post("/message/sms/send-global", payload=payload))

    def send_batch(self, *, messages: List[dict], from_whom: str = "4546", dispatch_id: int) -> Response:
        return Response(**self._request.post(
            "/message/sms/send-batch",
            payload={
                "messages": [
                    {
                        "user_sms_id": message["user_sms_id"],
                        "to": str(message["to"]),
                        "text": message["text"]
                    } for message in messages
                ],
                "from_whom": from_whom,
                "dispatch_id": dispatch_id
            }))

    def get_user_messages(self, from_date: str, to_date: str) -> Response:
        return Response(**self._request.get(
            "/message/sms/get-user-messages",
            payload={
                "from_date": from_date,
                "to_date": to_date,
                "user_id": self.user.id
            }
        ))

    def get_user_messages_by_dispatch(self, dispatch_id: int) -> Response:
        return Response(**self._request.get(
            "/message/sms/get-user-messages-by-dispatch",
            payload={
                "dispatch_id": dispatch_id,
                "user_id": self.user.id
            }))

    def get_dispatch_status(self, dispatch_id: int) -> Response:
        return Response(**self._request.get(
            "/message/sms/get-dispatch-status",
            payload={
                "dispatch_id": dispatch_id,
                "user_id": self.user.id
            }))

    def create_template(self, name: str, text: str) -> Response:
        return Response(**self._request.post(
            "/template",
            payload={
                "name": name,
                "text": text,
            }))

    def update_template(self, template_id: int, name: str, text: str) -> Response:
        return Response(**self._request.put(
            f"/template/{template_id}",
            payload={
                "name": name,
                "text": text,
            }
        ))

    def get_template(self, template_id: int) -> Response:
        return Response(**self._request.get(f"/template/{template_id}"))

    def get_templates(self) -> Response:
        return Response(**self._request.get("/template"))

    def totals(self, year: int) -> Response:
        return Response(**self._request.post(
            "/user/totals",
            payload={
                "year": year,
                "user_id": self.user.id
            }))

    def get_limit(self) -> Response:
        return Response(**self._request.get("/user/get-limit"))



from enum import Enum


class Status(str, Enum):
    TOKEN_INVALID = "token-invalid"


class Message(str, Enum):
    EXPIRED_TOKEN = "Expired token"
    INVALID_CREDENTIALS = "Неверный Email или пароль"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eskiz_sms']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.1,<0.24.0', 'python-dotenv>=0.21.0,<0.22.0']

setup_kwargs = {
    'name': 'eskiz-sms',
    'version': '0.2.1',
    'description': 'Package for eskiz.uz/sms',
    'long_description': '# eskiz-sms\n\neskiz-sms package for eskiz.uz/sms \n\n[![Downloads](https://pepy.tech/badge/eskiz-sms)](https://pepy.tech/project/eskiz-sms)\n[![Downloads](https://pepy.tech/badge/eskiz-sms/month)](https://pepy.tech/project/eskiz-sms)\n[![Downloads](https://pepy.tech/badge/eskiz-sms/week)](https://pepy.tech/project/eskiz-sms)\n\n> :warning: **Please use the latest version. In previous versions, there are a lot of mistakes, bugs**\n\n# Installation\n\n```\npip install eskiz_sms\n```\n\n# Quickstart\n\n```python\nfrom eskiz_sms import EskizSMS\n\nemail = "your_email@mail.com"\npassword = "your_password"\neskiz = EskizSMS(email=email, password=password)\neskiz.send_sms(\'998991234567\', \'message\', from_whom=\'4546\', callback_url=None)\n```\n\n### Using pre-saved token\n\n```python\nfrom eskiz_sms import EskizSMS\n\nyour_saved_token = \'eySomething9320\'\neskiz = EskizSMS(\'email\', \'password\')\neskiz.token.set(your_saved_token)\n\neskiz.send_sms(\'998901234567\', message=\'message\')\n```\n\n### Saving token to env file\n\nIf you set `save_token=True` it will save the token to env file\n\n```python\nfrom eskiz_sms import EskizSMS\n\neskiz = EskizSMS(\'email\', \'password\', save_token=True, env_file_path=\'.env\')\n# Don\'t forget to add env file to .gitignore!\nresponse = eskiz.send_sms(\'998901234567\', message=\'message\')\n```\n### Async usage\n\n```python\nimport asyncio\n\nfrom eskiz_sms.async_ import EskizSMS\n\n\nasync def main():\n    eskiz = EskizSMS(\'email\', \'password\')\n    response = await eskiz.send_sms(\'998901234567\', \'Hello, World!\')\n\n\nasyncio.run(main())\n```',
    'author': 'Malikov',
    'author_email': 'mlkv.dev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/malikovss/eskiz-sms',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from dotenv import get_key, set_key

from .logging import logger
from .request import BaseRequest

ESKIZ_TOKEN_KEY = "ESKIZ_TOKEN"


class Token(BaseRequest):
    __slots__ = (
        "auto_update",
        "save_token",
        "env_file_path",
        "_value",
        "_credentials",
        "updated_at",
        "__token_checked",
    )

    def __init__(
            self,
            email: str,
            password: str,
            save_token=False,
            env_file_path=None,
            auto_update: bool = True,
            is_async: bool = False,
    ):
        self._is_async = is_async
        self.auto_update = auto_update
        self.save_token = save_token

        self._value = None
        self._credentials = dict(email=email, password=password)

        if save_token:
            if env_file_path is None:
                env_file_path = '.env'
            self.env_file_path = env_file_path

        self.__token_checked = False

    def set(self, value):
        self._value = value

    def _save_to_env(self):
        set_key(self.env_file_path, key_to_set=ESKIZ_TOKEN_KEY,
                value_to_set=self._value)
        logger.info(f"Eskiz token saved to {self.env_file_path}")

    def _get_from_env(self):
        return get_key(dotenv_path=self.env_file_path, key_to_get=ESKIZ_TOKEN_KEY)

    def __str__(self):
        if self._value:
            return self._value
        return "None"

    __repr__ = __str__

    def _get(self, get_new: bool = False):
        if get_new or (
                not self.save_token and not self._value
        ) or (
                not self._value and self.save_token and not self._get_from_env()
        ):
            return self._get_new_token()

        if not self._value and self.save_token:
            self._value = self._get_from_env()
        self._check()
        return self._value

    def get(self, get_new: bool = False):
        if not get_new and self._value and self.__token_checked:
            return self._value

        if self._is_async:
            return self._a_get(get_new)
        return self._get(get_new)

    def _get_new_token(self) -> str:
        response = self._request(
            self._prepare_request(
                "POST",
                "/auth/login",
                self._credentials
            )
        )
        self._value = response.data['data']['token']
        self.__token_checked = True
        if self.save_token:
            self._save_to_env()
        return self._value

    def _check(self):
        self._request(
            self._prepare_request(
                "GET",
                "/auth/user",
                headers={
                    'Authorization': f'Bearer {self._value}'
                }
            )
        )

    # =====Async functions==== #
    async def _a_get(self, get_new: bool = False):
        if get_new or (
                not self.save_token and not self._value
        ) or (
                not self._value and self.save_token and not self._get_from_env()
        ):
            return await self._a_get_new_token()

        if not self._value and self.save_token:
            self._value = self._get_from_env()
        await self._a_check()
        return self._value

    async def _a_check(self):
        await self._a_request(
            self._prepare_request(
                "GET",
                "/auth/user",
                headers={
                    'Authorization': f'Bearer {self._value}'
                }
            )
        )

    async def _a_get_new_token(self):
        response = await self._a_request(
            self._prepare_request(
                "POST",
                "/auth/login",
                self._credentials
            )
        )
        self._value = response.data['data']['token']
        self.__token_checked = True
        if self.save_token:
            self._save_to_env()
        return self._value



from . import exceptions
from . import types
from .base import url_validator
from .eskiz import EskizSMS

__version__ = '0.2.1'

__all__ = [
    'EskizSMS',
    "exceptions",
    "types",
    "url_validator",
    "__version__",
]
